# -*- encoding: utf-8 -*-

$stdout.sync=true
$stderr.sync=true
libfilebase = 'rangesmaller.rb'
libfilepath = nil
er=nil
pathcand = %w(../lib/rangesmaller/ lib/rangesmaller/ rangesmaller/) + ['']
pathcand.each do |dir|
  begin
    s = dir+libfilebase
    require s
    libfilepath = s
    break
  rescue LoadError => er
  end
end
if libfilepath.nil?
  raise er
end

printf "NOTE: Library relative path: %s\n", libfilepath
print  "NOTE: Library full path: "
p $LOADED_FEATURES.grep(/#{Regexp.quote(libfilebase)}$/)

#################################################
# Unit Test
#################################################

#if $0 == __FILE__
  require 'minitest/unit'
  require 'minitest/autorun'
  # MiniTest::Unit.autorun

  # Taken from ((<URL:http://www.ruby-doc.org/core-2.1.1/Range.html>))
  class Xs                # represent a string of 'x's
    include Comparable
    attr :length
    def initialize(n)
      @length = n
    end
    def succ
      Xs.new(@length + 1)
    end
    def <=>(other)
      @length <=> other.length
    end
    def to_s
      sprintf "%2d #{inspect}", @length
    end
    def inspect
      'x' * @length
    end
  end

  # Used in test_bsearch_special
  class Special
    def [](f)
      (f>3.5 && f<4) ? true : false
    end
  end

  class TestFoo < MiniTest::Unit::TestCase
    def setup
      @ib = 1
      @ie = 6
      @r11 = (@ib..@ie)		# incl, incl
      @r12 = (@ib...@ie)	# incl, excl
      @s11 = Rangesmaller.new(@r11,  :exclude_begin => false)	# (1..6)   incl, incl
      @s21 = Rangesmaller.new(@r11,  :exclude_begin => true)	# (1<..6)  excl, incl
      @s12 = Rangesmaller.new(@r12,  :exclude_begin => false)	# (1...6)  incl, excl
      @s22 = Rangesmaller.new(@r12,  :exclude_begin => true)	# (1<...6) excl, excl
    end
    # teardown is not often used.
    def teardown
      @foo = nil
    end

    def test_new
      e22 = Rangesmaller.new(@s22)
      assert  e22.exclude_begin?
      assert  e22.exclude_end?
      assert_equal @ib, e22.begin
      assert_equal @ie, e22.end

      f12 = Rangesmaller.new(@r12)
      assert !f12.exclude_begin?
      assert  f12.exclude_end?
      assert_equal @ib, f12.begin
      assert_equal @ie, f12.end

      g11 = Rangesmaller.new(@ib, @ie)
      assert !g11.exclude_begin?
      assert !g11.exclude_end?
      assert_equal @ib, g11.begin
      assert_equal @ie, g11.end

      h12 = Rangesmaller.new(@ib, @ie, true)
      assert !h12.exclude_begin?
      assert  h12.exclude_end?
      assert_equal @ib, h12.begin
      assert_equal @ie, h12.end

      j22 = Rangesmaller.new(-@ie, -@ib, :exclude_end => true, :exclude_begin => true)
      assert  j22.exclude_begin?
      assert  j22.exclude_end?
      assert_equal(-@ie, j22.begin)
      assert_equal(-@ib, j22.end)
    end

    def test_new_const
      assert_equal @s22, Rangesmaller(@r12, :exclude_begin => true)
    end

    def test_exclude_begin
      b = 1
      e = 4
      r1 = (b...e)
      assert !Rangesmaller.new(r1).exclude_begin?
      assert  Rangesmaller.new(r1).exclude_end?
      assert  Rangesmaller.new(r1, :exclude_begin => 5).exclude_begin?
      assert  Rangesmaller.new(r1, :exclude_begin => 5).exclude_end?
      assert !Rangesmaller.new(b,e).exclude_begin?
      assert !Rangesmaller.new(b,e).exclude_end?
      assert !Rangesmaller.new(b,e,nil).exclude_begin?
      assert !Rangesmaller.new(b,e,nil).exclude_end?
      assert  Rangesmaller.new(b,e,'c').exclude_end?
      assert !Rangesmaller.new(b,e,:a =>5).exclude_begin?
      assert !Rangesmaller.new(b,e,:exclude_end => false, :exclude_begin => false).exclude_begin?
      assert !Rangesmaller.new(b,e,:exclude_end => false, :exclude_begin => false).exclude_end?
      assert !Rangesmaller.new(b,e,:exclude_end => true,  :exclude_begin => false).exclude_begin?
      assert  Rangesmaller.new(b,e,:exclude_end => true,  :exclude_begin => false).exclude_end?
      assert  Rangesmaller.new(b,e,:exclude_end => true,  :exclude_begin => true ).exclude_begin?
      assert  Rangesmaller.new(b,e,:exclude_end => true,  :exclude_begin => true ).exclude_end?
      assert  Rangesmaller.new(b,e,:exclude_end => false, :exclude_begin => true ).exclude_begin?
      assert !Rangesmaller.new(b,e,:exclude_end => false, :exclude_begin => true ).exclude_end?

      assert_raises ArgumentError do
        Rangesmaller.new()
      end
      assert_raises ArgumentError do
        Rangesmaller.new(5)
      end
      assert_raises ArgumentError do
        Rangesmaller.new(nil,5)
      end
      assert_raises ArgumentError do
        Rangesmaller.new(nil,5,true)
      end
      assert_raises ArgumentError do
        Rangesmaller.new(3,5,8,9)
      end
    end	# def test_exclude_begin

    # Test of Range#== because it has changed first!
    def test_eql_range
      assert       (@r11 == (@ib..@ie))
      assert_equal  @r11, (@ib..@ie)
      assert       (@r11 != @r12)
      assert_equal  @r12, (@ib...@ie)
      assert_equal  @r11, @s11
      assert       (@r11 != @s12)
      assert       (@r11 != @s21)
      assert       (@r11 != @s22)
      assert_equal  @r12, @s12
      assert       (@r12 != @s11)
      assert       (@r12 != @s21)
      assert       (@r12 != @s22)
      assert       (@r12 != 'K')
      assert       (@r12 != @ib)
    end	# def test_eql_range

    def test_eql
      t21 = Rangesmaller.new(@ib..@ie, :exclude_begin => true)
      u11 = Rangesmaller.new(3, @ie, :exclude_begin => false)
      v11 = Rangesmaller.new(@ib, 4, :exclude_begin => false)
      v21 = Rangesmaller.new(@ib..4, :exclude_begin => true)
      v22 = Rangesmaller.new(@ib...4,:exclude_begin => true)

      assert  (@s11 == @r11)
      assert_equal  @s11, @r11
      assert !(@s11 == @r12)
      assert  (@s11 != @r12)
      assert_equal  @s21, t21
      assert  (@s21 ==  t21)
      assert  (@s21 != @s11)
      assert  (@s21 != @s12)
      assert  (@s21 != @s22)
      assert  (@s12 == @r12)
      assert   @s12.eql?(@r12)
      assert  (@s11 != u11)
      assert  (@s11 != v11)
      assert  (@s21 != v21)
      assert  (@s22 != v22)
      assert  (@s22 != u11)
      assert  (@s22 != 'a')

      sc1 = Rangesmaller.new("D".."F", :exclude_begin => false)
      sc2 = Rangesmaller.new("D".."F", :exclude_begin => true)
      assert (sc1 == ("D".."F"))
      assert (sc1 == Rangesmaller.new(("D".."F")))
      assert (sc1 != sc2)
    end	# def test_eql

    def test_eql2
      assert  ((1..2) == (1..2))
      assert  ((1..2) == (1.0..2))
      assert  ((1..2).eql? (1..2))	# 1==(1.0)     == true
      assert !((1..2).eql? (1.0..2))	# 1.eql?(1.0)  == false
      r12 = Rangesmaller(1..2, :exclude_begin => true)
      assert  (r12.eql? Rangesmaller(1..2, :exclude_begin => true))
      assert !(r12.eql? Rangesmaller(1.0..2, :exclude_begin => true))
    end

    def test_begin
      assert_equal(@ib, @s11.begin)
      assert_equal(@ib, @s12.begin)
      assert_equal(@ib, @s21.begin)
      assert_equal(@ib, @s22.begin)
    end

    def test_end
      assert_equal(@ie, @s11.end)
      assert_equal(@ie, @s12.end)
      assert_equal(@ie, @s21.end)
      assert_equal(@ie, @s22.end)
    end

    def test_first
      # This is the key method of this class.

      # irb> (5...8.9).last(1)	# => [8]
      # irb> (5.2..9).last(1)	# => TypeError: can't iterate from Float
      # irb> (5.2..9).first(1)	# => TypeError: can't iterate from Float
      # irb> (5.2...9).first(1)	# => TypeError: can't iterate from Float
      # irb> Rangesmaller(5.2,9,:exclude_begin=>true).first(1)	# => [6]	# Not right!
      #

      assert_equal(@ib,   @s11.first)
      assert_equal(@ib,   @s12.first)
      assert_equal(@ib,   @s21.first)
      assert_equal(@ib,   @s22.first)
      assert_equal(@ib,   @s12.first(1)[0])
      assert_equal(@ib+1, @s21.first(1)[0])
      assert_equal(@ib+1, @s22.first(1)[0])

      assert_equal(nil, Rangesmaller.new(9, 3, :exclude_begin => true).first(1)[0])

      assert_raises ArgumentError do
        @s22.first(1, 3)
      end

      ## String
      sc1 = Rangesmaller.new("D".."F", :exclude_begin => false)
      sc2 = Rangesmaller.new("D".."F", :exclude_begin => true)
      assert_equal('D',  sc1.first)
      assert_equal('D',  sc2.first)
      assert_equal('E',  sc2.first(1)[0])

      ## Arbitrary Class
      sx1 = Rangesmaller.new((Xs.new(3)..Xs.new(6)), :exclude_begin => true)
      assert_equal(Xs.new(3), sx1.first)
      assert_equal(Xs.new(4), sx1.first(1)[0])

      ## Float
      sf1 = Rangesmaller.new((-1.4)..8, :exclude_begin => false)
      sf2 = Rangesmaller.new((-1.4)..8, :exclude_begin => true)
      assert_equal(-1.4,  sf1.first)
      assert_equal(-1.4,  sf2.first)
      assert_raises TypeError do
        sf1.first(1)
      end
      assert_raises TypeError do
        sf2.first(1)
      end

      ## Else
      sn = Rangesmaller.new(nil..nil, :exclude_begin => true)
      assert_raises TypeError do
        sn.first(1)[0]
      end

      assert_raises ArgumentError do
        @s22.first(-7)	# "negative array size (or size too big)"
      end
      assert_raises TypeError do
        @s22.first('a')
      end
      assert_raises TypeError do
        @s22.first(nil)
      end
    end	# def test_first


    def test_each
      ns=0; @s11.each{|i| ns+=i}
      assert_equal(@r11.reduce(:+), ns)
      ns=0; @s12.each{|i| ns+=i}
      assert_equal(@r12.reduce(:+), ns)
      ns=0; @s21.each{|i| ns+=i}
      assert_equal(((@ib+1)..@ie).reduce(:+), ns)
      ns=0; ret=@s22.each{|i| ns+=i}
      assert_equal(((@ib+1)...@ie).reduce(:+), ns)
      assert_equal(@s22, ret)
      assert(Enumerator === @s22.each)

      ## Arbitrary Class
      sx1 = Rangesmaller.new(Xs.new(3), Xs.new(6), :exclude_begin => true, :exclude_end => true)
      a=[]; sx1.each{|i| a.push(i)}
      assert_equal([Xs.new(4), Xs.new(5)], a)
    end	# def test_each


    def test_last	# Apparently it uses each() internally.
      assert_equal(@ie,   @s11.last)
      assert_equal(@ie-1, @s12.last(1)[0])
      assert_equal(@ie,   @s21.last)
      assert_equal(@ie-1, @s22.last(1)[0])
    end


    def test_include
      assert  @s11.include?(@ib)
      assert  @s11.include?(@ie)
      assert  @s12.include?(@ib)
      assert !@s12.include?(@ie)
      assert  @s12.include?(@ie-1)
      assert !@s21.include?(@ib)
      assert  @s21.include?(@ib+1)
      assert  @s21.include?(@ie)
      assert !@s22.include?(@ib)
      assert !@s22.include?(@ie)
      assert  (@s11 === @ib)
      assert !(@s21 === @ib)
      assert  (@s21 === @ib+1)
      assert !(@s22 === @ie)

      assert  (Rangesmaller.new("a", "z") === "c")
      assert !(Rangesmaller.new("a", "z") === "cc")	# Key! (see cover?)
      assert  (@s22 === (@ib+@ie)/2.0+0.1)	# Key! (see cover?)
      assert  (Rangesmaller.new("a", "z").member?("c"))
      assert !(Rangesmaller.new("a", "z").member?("cc"))
      assert !(Rangesmaller.new("a", "z", :exclude_begin => 777) === "a")
      assert  (Rangesmaller.new("a", "z", :exclude_begin => nil) === "a")
      assert  (Rangesmaller.new("a", "z", :exclude_begin => 777) === "b")
    end	# def test_include


    def test_bsearch
      ary = [0, 4, 7, 10, 12]
      assert_equal(2,   Rangesmaller(0, ary.size).bsearch{|i| ary[i] >= 6})
      # http://www.ruby-doc.org/core-2.1.1/Range.html#method-i-bsearch

      assert_equal(nil, Rangesmaller(3...4).bsearch{  |i| ary[i] >= 11})
      assert_equal(nil, Rangesmaller(3.6...4).bsearch{|i| ary[i] >= 11})
      assert_equal(4,   Rangesmaller(3...5).bsearch{  |i| ary[i] >= 11})
      assert_equal(4.0, Rangesmaller(3...5.1).bsearch{|i| ary[i] >= 11})
      assert_equal(nil, Rangesmaller(3.6...4).bsearch{|i| ary[i] >= 11})

      assert_equal(4,   Rangesmaller(3...5,   :exclude_begin => 1).bsearch{  |i| ary[i] >= 11})
      assert_equal(nil, Rangesmaller(4...5,   :exclude_begin => 1).bsearch{  |i| ary[i] >= 11})
      assert_equal(4.0, Rangesmaller(3...5.1, :exclude_begin => 1).bsearch{|i| ary[i] >= 11})
      assert_equal(nil, Rangesmaller(3.6...4, :exclude_begin => 1).bsearch{|i| ary[i] >= 11})

      assert_raises TypeError do
        Rangesmaller.new((?a..?b), :exclude_begin => true).bsearch{|i| ary[i] >= 11}
      end
    end	# def test_bsearch

    def test_bsearch_special
      sp = Special.new

      # Standard Range
      assert_equal(nil, Rangesmaller(3..4).bsearch{ |i| sp[i]})
      assert_equal(nil, Rangesmaller(3...4).bsearch{|i| sp[i]})
      assert(1e-8 > (Rangesmaller(3.0...4).bsearch{|i| sp[i]} - 3.5).abs)
      assert(1e-8 > (Rangesmaller(3...4.0).bsearch{|i| sp[i]} - 3.5).abs)
      assert(1e-8 > (Rangesmaller(3.3..4).bsearch{ |i| sp[i]} - 3.5).abs)
 
      # Rangesmaller
      assert(1e-8 > (Rangesmaller(3...4.1,   :exclude_begin => 1).bsearch{|i| sp[i]} - 3.5).abs)
      assert(1e-8 > (Rangesmaller(3.7...4,   :exclude_begin => 1).bsearch{|i| sp[i]} - 3.7).abs)
      assert(1e-8 > (Rangesmaller(3.7...4.2, :exclude_begin => 1).bsearch{|i| sp[i]} - 3.7).abs)	# If end is 4.7, it will be nil (presumably due to the algorithm), whereas still 3.7 if 4.5.
    end	# def test_bsearch_special


    def test_cover
      assert  @s11.cover?(@ib)
      assert  @s11.cover?(@ie)
      assert  @s12.cover?(@ib)
      assert !@s12.cover?(@ie)
      assert  @s12.cover?(@ie-1)
      assert !@s21.cover?(@ib)
      assert  @s21.cover?(@ib+1)
      assert  @s21.cover?(@ie)
      assert !@s22.cover?(@ib)
      assert !@s22.cover?(@ie)
      st = Rangesmaller.new((1.4..7), :exclude_begin => true)
      assert !(st.cover?(1.4))
      assert  (st.cover?(1.5))

      assert  (Rangesmaller.new("a", "z").cover?("c"))
      assert  (Rangesmaller.new("a", "z").cover?("cc"))	# Key! (see include?)
      assert  (@s22.cover?((@ib+@ie)/2.0+0.1))	# Key! (see cover?)
      su = Rangesmaller.new("a", "z", :exclude_begin => 777)
      sv = Rangesmaller.new("a", "z", :exclude_begin => nil)
      assert !(su.cover?("a"))
      assert  (sv.cover?("a"))
      assert  (su.cover?("b"))
    end	# def test_cover


    def test_hash
      assert_equal(@r11.hash, @s11.hash)
      assert_equal(@r12.hash, @s12.hash)
      assert (@r12.hash != @s22.hash)
    end	# def test_hash


    def test_min
      assert_equal(@ib,   @s11.min)
      assert_equal(@ib,   @s12.min)
      assert_equal(@ib+1, @s21.min)
      assert_equal(@ib+1, @s22.min)

      assert_equal(@ie-1, @s22.min{|a,b| -a <=> -b})

      assert_raises TypeError do
        Rangesmaller.new(1.0, 5, :exclude_begin => true).min
      end
    end	# def test_min


    def test_min_by
      assert_equal(@ib+1, @s22.min_by.each{|i| i})

      assert_raises TypeError do
        Rangesmaller.new(1.0, 5, :exclude_begin => true).min_by
      end

      assert_equal(@ie-1, @s22.min_by{|a| -a })
    end	# def test_min_by


    def test_minmax
      assert_equal([@ib,@ie],     @s11.minmax)
      assert_equal([@ib,@ie-1],   @s12.minmax)
      assert_equal([@ib+1,@ie],   @s21.minmax)
      assert_equal([@ib+1,@ie-1], @s22.minmax)

      assert_equal([@ie-1,@ib+1], @s22.minmax{|a,b| -a <=> -b})	# Not the best test...

      assert_raises TypeError do
        Rangesmaller.new(1.0, 5, :exclude_begin => true).minmax
      end
    end	# def test_minmax


    def test_minmax_by
      assert_equal([@ib+1,@ie-1], @s22.minmax_by.each{|i| i})

      assert_raises TypeError do
        Rangesmaller.new(1.0, 5, :exclude_begin => true).minmax_by
      end

      assert_equal([@ie-1,@ib+1], @s22.minmax_by{|a| -a })
    end	# def test_minmax_by


    def test_size
      assert_equal(@ie-@ib+1, @s11.size)
      assert_equal(@ie-@ib,   @s12.size)
      assert_equal(@ie-@ib,   @s21.size)
      assert_equal(@ie-@ib-1, @s22.size)
      Rangesmaller.new(9, -5, :exclude_begin => true)
      assert_equal 0, Rangesmaller.new(9, -5, :exclude_begin => true).size
      assert_equal nil, Rangesmaller.new("a", "c").size
      assert_equal nil, Rangesmaller.new(nil, nil).size

      # Infinity
      inf = Float::INFINITY
      excl_ini = { :exclude_begin => true }
      assert_equal inf, Rangesmaller.new(-inf, 1).size
      assert_equal   0, Rangesmaller.new(-inf, -inf, excl_ini).size
      assert_equal inf, Rangesmaller.new(-inf,    1, excl_ini).size
      assert_equal   0, Rangesmaller.new( inf,  inf, excl_ini).size
      assert_equal inf, Rangesmaller.new(   1,  inf, excl_ini).size

      # Float
      rfi = (2.4..4.4)	# size() => 3	see [ruby-list:49797] from matz
      rfe = (2.4...4.4)	# size() => 2
      siz = rfi.size
      assert_equal siz,   Rangesmaller(rfi).size
      assert_equal siz-1, Rangesmaller(rfe).size
      assert_equal siz-1, Rangesmaller(rfi, excl_ini).size
      assert_equal siz-2, Rangesmaller(rfe, excl_ini).size
      assert_equal siz-1, Rangesmaller(Rational(24,10)..4.4,  excl_ini).size
      assert_equal siz-2, Rangesmaller(Rational(24,10)...4.4, excl_ini).size
      # (0.5...5).size			# => 5 (Ruby 2.1)
      # (Rational(1,2)...5).size	# => 4 (Ruby 2.1) =>  Bug!

      # String
      rsi = (?a..?d)
      if rsi.size.nil?
        assert_equal nil, Rangesmaller(rsi, excl_ini).size	# Ruby 2.1
      else
        assert_equal   3, Rangesmaller(rsi, excl_ini).size	# If the specification ever changes?
      end
    end	# def test_size


    def test_step
      ns=0; @s11.step(2){|i| ns+=i}
      assert_equal(1+3+5, ns)
      ns=0; @s12.step(2){|i| ns+=i}
      assert_equal(1+3+5, ns)
      ns=0; @s21.step(2){|i| ns+=i}
      assert_equal(2+4+6, ns)
      ns=0; ret=@s22.step(2){|i| ns+=i}
      assert_equal(2+4, ns)
      assert_equal(@s22, ret)
      assert(Enumerator === @s22.step(2))

      ## Arbitrary Class
      sx1 = Rangesmaller.new(Xs.new(3), Xs.new(6), :exclude_begin => true, :exclude_end => true)
      a=[]; sx1.step(2){|i| a.push(i)}
      assert_equal([Xs.new(4)], a)
    end	# def test_step


    def test_to_s
      assert_equal 'a...c',  Rangesmaller(?a...?c).to_s
      assert_equal 'a<..c',  Rangesmaller(?a..?c,  :exclude_begin => true).to_s
      assert_equal 'a<...c', Rangesmaller(?a...?c, :exclude_begin => true).to_s
      assert_equal '"a"<.."c"',  Rangesmaller(?a..?c,  :exclude_begin => true).inspect
    end

  end	# class TestFoo < MiniTest::Unit::TestCase

#end	# if $0 == __FILE__


# % ruby1.8 -rlib/rangesmaller/rangesmaller.rb -e 'p Rangesmaller(2.3...7,:exclude_begin=>1).first==2.3'
# ruby1.8 -rlib/rangesmaller/rangesmaller.rb -e 'i=0;p Rangesmaller(1...4,:exclude_begin=>true).each{|j|i+=j};p i==5'
# ruby1.8 -rlib/rangesmaller/rangesmaller.rb -e 'p Rangesmaller(-1...7,:exclude_begin=>1).min==0'          
